/*
    nd version 1.0
    Copyright (C) 2008 Andrew Madigan

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along
    with this program; if not, write to the Free Software Foundation, Inc.,
    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
*/
#ifndef _POSIX_C_SOURCE
#define _POSIX_C_SOURCE 200112L
#endif

#ifndef _XOPEN_SOURCE
#define _XOPEN_SOURCE 600
#endif

#include <unistd.h>
#include <sys/socket.h>
#include <netdb.h>
#include <string.h>
#include <errno.h>
#include <netinet/in.h>
#include <stdbool.h>
#include <stdlib.h>
#include <fcntl.h>
#include <stdio.h>
#include <signal.h>
#include <arpa/inet.h>
#include <sys/wait.h>
#include <stdint.h>
#include <pwd.h>
#include <time.h>

void nd_print_help(FILE* out, char* name);
void nd_child_died(int signal, siginfo_t* info, void* context);
void nd_setsignals();
void nd_exec(int sockfd, bool ignore_pipe, struct sockaddr_in sockaddr,
             const char* srvportstr, const char* srvaddrstr, 
             char* const argv[]);
void nd_setlog(const char* logfile);
void nd_writepid(const char* pidfile);
void nd_daemonize();
int nd_open_socket(const char* addr, uint16_t port);
void nd_stopsig(int signal);

int verbosity = 0;
time_t time_ms;
struct tm* time_full = NULL;
bool kill_children = true;

int main(int argc, char* argv[]) {
   char* srvportstr = NULL;
   char* srvaddrstr = "0.0.0.0";
   bool daemonize = false;
   bool ignore_pipe = false;
   char* logfile = NULL;
   char* pidfile = NULL;
   char* user = NULL;
   bool setserveruid = false;
   
   int ch;
   
   while ((ch = getopt(argc, argv, ":p:a:vdil:x:u:U:hk")) != -1) {
      switch (ch) {
      case 'p' :
         srvportstr = optarg;
         break;
      case 'a' :
         srvaddrstr = optarg;
         break;
      case 'v' :
         if (verbosity < 9) verbosity++;
         break;
      case 'd' :
         daemonize = true;
         break;
      case 'i' :
         ignore_pipe = true;
         break;
      case 'l' :
         logfile = optarg;
         break;
      case 'x' :
         pidfile = optarg;
         break;
      case 'U' :
         setserveruid = true;
      case 'u' :
         user = optarg;
         break;
      case 'h' :
         nd_print_help(stdout, argv[0]);
         return 0;
      case 'k':
         kill_children = false;
         break;
      }
   }
   
   
   
   //Setup logfile and pid file.
   if (logfile != NULL) {
      nd_setlog(logfile);
   }
   
   if (verbosity > 0) {
      time_ms = time(NULL);
      time_full = localtime(&time_ms);
      fprintf(stderr, "[%02d:%02d:%02d %02d/%02d/%04d] Server start, pid %d\n", 
               time_full->tm_hour, time_full->tm_min, time_full->tm_sec, 
               time_full->tm_mon + 1, time_full->tm_mday, time_full->tm_year 
               + 1900, getpid());
   }
   
   if (srvportstr == NULL) {
      fprintf(stderr, "Error: You must pass a port number.\n\n");
      nd_print_help(stderr, argv[0]);
      return 64; //EX_USAGE
   }
   
   errno = 0;
   long portl = strtol(srvportstr, (char**) NULL, 10);
   if (errno != 0) {
      fprintf(stderr, "Error: Unable to parse port number %s, %s\n", srvportstr, strerror(errno));
      return 64;
   }
   
   if (portl > UINT16_MAX || portl < 0) {
      fprintf(stderr, "Error: %ld is not a valid port number.\n", portl);
      return 64;
   }
   
   //We have the server socket, daemonize and setuid (if needed)
   if (daemonize) {
      nd_daemonize();
      if (verbosity > 0) {
         time_ms = time(NULL);
         time_full = localtime(&time_ms);
         fprintf(stderr, "[%02d:%02d:%02d %02d/%02d/%04d] Server daemonized, new pid %d\n", 
                  time_full->tm_hour, time_full->tm_min, time_full->tm_sec, 
                  time_full->tm_mon + 1, time_full->tm_mday, time_full->tm_year 
                  + 1900, getpid());
      }
   }
   
   nd_setsignals();
   
   if (pidfile != NULL) {
      nd_writepid(pidfile);
   }
   
   int newuid = -1;
   if (user != NULL) {
      struct passwd* info = getpwnam(user);
      if (info == NULL) {
         fprintf(stderr, "Error: Unable to find user %s, %s\n", user, strerror(errno));
         return 67; //EX_NOUSER
      }
      newuid = info->pw_uid;
   }
   
   int srvfd = nd_open_socket(srvaddrstr, (uint16_t) portl);
   if (verbosity > 0) {
      time_ms = time(NULL);
      time_full = localtime(&time_ms);
      fprintf(stderr, "[%02d:%02d:%02d %02d/%02d/%04d] Socket open, address: %s, port: %s\n", 
               time_full->tm_hour, time_full->tm_min, time_full->tm_sec, 
               time_full->tm_mon + 1, time_full->tm_mday, time_full->tm_year 
               + 1900, srvaddrstr, srvportstr);
   }
   
   if (setserveruid) {
      if (setuid(newuid) == -1) {
         fprintf(stderr, "Unable to setuid %d (%s) for server: %s\n", newuid, user, strerror(errno));
         return 71;
      }
      if (verbosity > 0) {
         time_ms = time(NULL);
         time_full = localtime(&time_ms);
         fprintf(stderr, "[%02d:%02d:%02d %02d/%02d/%04d] Changed server user to %s (UID: %d)\n", 
                  time_full->tm_hour, time_full->tm_min, time_full->tm_sec, 
                  time_full->tm_mon + 1, time_full->tm_mday, time_full->tm_year 
                  + 1900, user, newuid);
      }
   }
   
   struct sockaddr_in sockaddr;
   socklen_t sockaddrlen;
   
   if (verbosity > 0) {
      time_ms = time(NULL);
      time_full = localtime(&time_ms);
      fprintf(stderr, "[%02d:%02d:%02d %02d/%02d/%04d] Entering socket accept loop\n", 
                  time_full->tm_hour, time_full->tm_min, time_full->tm_sec, 
                  time_full->tm_mon + 1, time_full->tm_mday, time_full->tm_year 
                  + 1900);
   }
   while (true) {
      if (verbosity > 2) {
         time_ms = time(NULL);
         time_full = localtime(&time_ms);
         fprintf(stderr, "[%02d:%02d:%02d %02d/%02d/%04d] Waiting for connections...\n", 
                     time_full->tm_hour, time_full->tm_min, time_full->tm_sec, 
                     time_full->tm_mon + 1, time_full->tm_mday, time_full->tm_year 
                     + 1900);
      }
      sockaddrlen = sizeof(sockaddr);
      int sockfd = accept(srvfd, (struct sockaddr*) &sockaddr, &sockaddrlen);
      if (sockfd == -1) {
         if (errno == EINTR) continue;
         fprintf(stderr, "Unable to accept connection on port %s: %s\n", srvportstr, strerror(errno));
         return 74;
      }
      if (verbosity > 1) {
         char port[6];
         sprintf(port, "%d", ntohs(sockaddr.sin_port));
         char addrstr[INET_ADDRSTRLEN];
         in_addr_t addr = sockaddr.sin_addr.s_addr;
         inet_ntop(AF_INET, &addr, addrstr, INET_ADDRSTRLEN * sizeof(char));
         time_ms = time(NULL);
         time_full = localtime(&time_ms);
         fprintf(stderr, "[%02d:%02d:%02d %02d/%02d/%04d] Connection received from %s, port %s\n", 
                     time_full->tm_hour, time_full->tm_min, time_full->tm_sec, 
                     time_full->tm_mon + 1, time_full->tm_mday, time_full->tm_year 
                     + 1900, addrstr, port);
      }
      pid_t handler_pid;
      switch ((handler_pid = fork())) {
      case 0:
         //Child process
         if (user != NULL) {
            if (setuid(newuid) == -1) {
               fprintf(stderr, "Unable to setuid %d (%s) for child: %s\n", newuid, user, strerror(errno));
               return 71;
            }
            if (verbosity > 1) {
               time_ms = time(NULL);
               time_full = localtime(&time_ms);
               fprintf(stderr, "[%02d:%02d:%02d %02d/%02d/%04d] Handler %d: switched to user %s (UID: %d) \n", 
                           time_full->tm_hour, time_full->tm_min, time_full->tm_sec, 
                           time_full->tm_mon + 1, time_full->tm_mday, time_full->tm_year 
                           + 1900, getpid(), user, newuid);
            }
         }
         
         nd_exec(sockfd, ignore_pipe, sockaddr, srvportstr, srvaddrstr, &argv[optind]);
         
         break;
      case -1:
         fprintf(stderr, "Unable fork child process: %s\n", strerror(errno));
      default: 
         close(sockfd);
         if (verbosity > 1) {
            char port[6];
            sprintf(port, "%d", ntohs(sockaddr.sin_port));
            char addrstr[INET_ADDRSTRLEN];
            in_addr_t addr = sockaddr.sin_addr.s_addr;
            inet_ntop(AF_INET, &addr, addrstr, INET_ADDRSTRLEN * sizeof(char));
            time_ms = time(NULL);
            time_full = localtime(&time_ms);
            fprintf(stderr, "[%02d:%02d:%02d %02d/%02d/%04d] Handler forked, pid: %d\n", 
                        time_full->tm_hour, time_full->tm_min, time_full->tm_sec, 
                        time_full->tm_mon + 1, time_full->tm_mday, time_full->tm_year 
                        + 1900, handler_pid);
         }
      }
   }
   
}

void nd_setsignals() {
   struct sigaction action;
   sigemptyset(&action.sa_mask);
   action.sa_flags = SA_NOCLDSTOP | SA_SIGINFO;
   action.sa_sigaction = nd_child_died;
   sigaction(SIGCHLD, &action, NULL);
   action.sa_flags = 0;
   action.sa_handler = nd_stopsig;
   sigaction(SIGINT, &action, NULL);
   sigaction(SIGTERM, &action, NULL);
}

void nd_exec(int sockfd, bool ignore_pipe, struct sockaddr_in sockaddr,
             const char* srvportstr, const char* srvaddrstr, char* const argv[])
{
   if (ignore_pipe) signal(SIGPIPE, SIG_IGN);
   char verbose[2];
   verbose[0] = '0' + verbosity;
   verbose[1] = '\0';
   char port[6];
   sprintf(port, "%d", ntohs(sockaddr.sin_port));
   char addrstr[INET_ADDRSTRLEN];
   in_addr_t addr = sockaddr.sin_addr.s_addr;
   inet_ntop(AF_INET, &addr, addrstr, INET_ADDRSTRLEN * sizeof(char));
   setenv("ND_PORT", port, 1);
   setenv("ND_ADDR", addrstr, 1);
   setenv("ND_LISTEN_PORT", srvportstr, 1);
   setenv("ND_LISTEN_ADDR", srvaddrstr, 1);
   setenv("ND_VERBOSITY", verbose, 1);
   //Set the socket to stdin and stdout
   dup2(sockfd, 0);
   dup2(sockfd, 1);
   if (verbosity > 1) {
      time_ms = time(NULL);
      time_full = localtime(&time_ms);
      fprintf(stderr, "[%02d:%02d:%02d %02d/%02d/%04d] Handler %d environment setup, executing %s\n", 
                  time_full->tm_hour, time_full->tm_min, time_full->tm_sec, 
                  time_full->tm_mon + 1, time_full->tm_mday, time_full->tm_year 
                  + 1900, getpid(), argv[0]);
   }
   if (execvp(argv[0], argv) == -1) {
      fprintf(stderr, "Unable to execute child %s: %s\n", argv[0], strerror(errno));
      exit(64);
   }
}

void nd_writepid(const char* pidfile) {
   FILE* pidstrm = fopen(pidfile, "w");
   if (pidstrm == NULL) {
      fprintf(stderr, "Unable to open pid file %s: %s\n", pidfile, strerror(errno));
      exit(74);
   }
   if (fprintf(pidstrm, "%d\n", getpid()) < 0) {
      fprintf(stderr, "Unable to write pid to file %s: %s\n", pidfile, strerror(errno));
      exit(74);
   }
   
   if (fclose(pidstrm) == EOF) {
      fprintf(stderr, "Unable to close pid file %s: %s\n", pidfile, strerror(errno));
      exit(74);
   }
}

void nd_setlog(const char* logfile) {
   int logfd = open(logfile, O_WRONLY | O_CREAT | O_APPEND | O_NOCTTY);
   if (logfd == -1) {
      fprintf(stderr, "Unable to open log file %s: %s\n", logfile, strerror(errno));
      exit(74); //EX_IOERR
   }
   dup2(logfd, 2);
}

void nd_daemonize() {
   switch (fork()) {
   case 0: break;
   case -1:
      fprintf(stderr, "Unable to fork for daemonization: %s\n", strerror(errno));
      exit(71); //EX_OSERR
   default: exit(0);
   }
   
   if (setsid() < 0) {
      fprintf(stderr, "Unable to setsid for daemonization: %s\n", strerror(errno));
      exit(71);
   }
   
   switch (fork()) {
   case 0: break;
   case -1:
      fprintf(stderr, "Unable to double fork for daemonization: %s\n", strerror(errno));
      exit(71);
   default: exit(0);
   }
   
   close(0);
   close(1);
   if (isatty(2)) close(2);
   if (setpgid(0, 0) < 0) {
      fprintf(stderr, "Unable to setpgid for daemonization: %s\n", strerror(errno));
      exit(71);
   }
}

int nd_open_socket(const char* addr, uint16_t port) {
   struct sockaddr_in srvaddr;
   memset(&srvaddr, 0, sizeof(struct sockaddr_in));
   srvaddr.sin_family = AF_INET;
   srvaddr.sin_port = htons(port);
   srvaddr.sin_addr.s_addr = htonl(0);
   
   switch(inet_pton(AF_INET, addr, &srvaddr.sin_addr.s_addr)) {
   case 0:
      fprintf(stderr, "Error: Invalid address: %s\n", addr);
      exit(64);
   case -1:
      fprintf(stderr, "Error: Unable to parse address: %s, %s\n", addr, strerror(errno));
      exit(64);
   }
   
   struct protoent* tcp_ent = getprotobyname("tcp");
   if (tcp_ent == NULL) {
      fprintf(stderr, "Unable to find protocol tcp\n");
      exit(71);
   }
   int tcp = tcp_ent->p_proto;
   endprotoent();
   
   int srvfd = socket(AF_INET, SOCK_STREAM, tcp);
   if (srvfd == -1) {
      fprintf(stderr, "Error creating socket: %s\n", strerror(errno));
      exit(64);
   }
   fcntl(srvfd, F_SETFL, 0); //Make sure the server-socket is blocking
   fcntl(srvfd, F_SETFD, FD_CLOEXEC); //Don't pass the fd beyond to exec'd processes
   static const int flag = 1;
   setsockopt(srvfd, SOL_SOCKET, SO_REUSEADDR, &flag, sizeof(int));
   
   if (bind(srvfd, (struct sockaddr*) &srvaddr, sizeof(struct sockaddr_in)) == -1) {
      fprintf(stderr, "Unable to bind to port: %s\n", strerror(errno));
      exit(64);
   }
   
   int backlog = 10;
   if (backlog > SOMAXCONN) backlog = SOMAXCONN;
   if (listen(srvfd, backlog) == -1) {
      fprintf(stderr, "Unable to listen on port: %s\n", strerror(errno));
      exit(64);
   }
   
   return srvfd;
}

void nd_print_help(FILE* out, char* name) {
   fprintf(out, "Usage: %s -p port [OPTIONS]... command...\n", name);
   fprintf(out, "Listen on a TCP port, executing a command with stdin and stdout redirecting to the incoming connection.\n");
   fprintf(out, "\n");
   fprintf(out, "Option Summary:\n");
   fprintf(out, "\t-p port\t\tThe port to run the server on.\n");
   fprintf(out, "\t-a addr\t\tThe address to ind the server to.\n");
   fprintf(out, "\t-v\t\tPrint additional output.\n");
   fprintf(out, "\t-d\t\tDaemonize the server.\n");
   fprintf(out, "\t-i\t\tBlock SIGPIPE for child processes.\n");
   fprintf(out, "\t-l logfile\tRedirect output to logfile.\n");
   fprintf(out, "\t-x pidfile\tWrite out the PID of the server to pidfile.\n");
   fprintf(out, "\t-u user\t\tRun child processes as user.\n");
   fprintf(out, "\t-U user\t\tAfter binding to the address, switch to user.\n");
   fprintf(out, "\t-k\t\tDo not kill child processes when server is killed.\n");
   fprintf(out, "\t-h\t\tPrint this help.\n");
   fprintf(out, "\n");
   fprintf(out, "Homepage: <http://amadigan.sixbit.org>\n");
   fprintf(out, "Bug Reports: <amadigan@gmail.com>\n");
}

void nd_child_died(int signal, siginfo_t* info, void* context) {
   int stat_val;
   waitpid(info->si_pid, &stat_val, WNOHANG);
   if (info->si_status != 0) {
      if (info->si_errno != 0) {
         fprintf(stderr, "Child process %d died with exit code %d: %s\n", info->si_pid, info->si_status, strerror(info->si_status));
      }
      else {
         fprintf(stderr, "Child process %d died with exit code %d\n", info->si_pid, info->si_status);
      }
   }
   else if (verbosity > 1) {
      time_ms = time(NULL);
      time_full = localtime(&time_ms);
      fprintf(stderr, "[%02d:%02d:%02d %02d/%02d/%04d] Handler %d execution complete\n", 
                  time_full->tm_hour, time_full->tm_min, time_full->tm_sec, 
                  time_full->tm_mon + 1, time_full->tm_mday, time_full->tm_year 
                  + 1900, info->si_pid);
   }
}

void nd_stopsig(int signal) {
   if (kill_children) killpg(getpgid(0), signal);
   if (verbosity > 0) {
      time_ms = time(NULL);
      time_full = localtime(&time_ms);
      fprintf(stderr, "[%02d:%02d:%02d %02d/%02d/%04d] Shutting down server on signal %d\n", 
                  time_full->tm_hour, time_full->tm_min, time_full->tm_sec, 
                  time_full->tm_mon + 1, time_full->tm_mday, time_full->tm_year 
                  + 1900, signal);
   }
   fclose(stderr);
   exit(0);
}
